@@ -12,6 +12,8 @@
 #include <readline/history.h>
 #include <oniguruma.h>
 #include <sys/stat.h>
+#include <sys/types.h>
+#include <pwd.h>
 
 #include "config.h"
 #include "saphire_extra.h"
@@ -191,21 +193,18 @@ void saphire_get_quoted_fname_readline_ver(char* file, string_obj* ret)
 }
 
 
-char* readline_file_completion_migemo_match(char* editing_dir, char* text, char* file)
+char* readline_file_completion_migemo_match(char* text, char* file)
 {
-//printf("editing_dir (%s) text (%s) file (%s)\n", editing_dir, text, file);
-    /// textとfileの同じところまでは比較しない ///
-    int same_point = str_headsame(text+ strlen(editing_dir), file);
+    int same_point = str_headsame(text, file);
 
     OnigUChar * p;
     if(same_point == 0) {
-        p = migemo_query(gMigemo, text + strlen(editing_dir));
+        p = migemo_query(gMigemo, text);
     }
     /// text側がfileの部分なら判定するまでも無くマッチ ///
-    else if(same_point == strlen(text + strlen(editing_dir))) {
+    else if(same_point == strlen(text)) {
         char tmp[PATH_MAX];
-        strcpy(tmp, editing_dir);
-        strcat(tmp, file);
+        strcpy(tmp, file);
         string_obj* tmp2 = STRING_NEW("");
         saphire_get_quoted_fname_readline_ver(tmp, tmp2);
         char* ret = strdup(string_c_str(tmp2));
@@ -214,7 +213,7 @@ char* readline_file_completion_migemo_match(char* editing_dir, char* text, char*
         return ret;
     }
     else {
-        p = migemo_query(gMigemo, text + strlen(editing_dir)+same_point);
+        p = migemo_query(gMigemo, text+same_point);
     }
 
 
@@ -279,8 +278,7 @@ char* readline_file_completion_migemo_match(char* editing_dir, char* text, char*
 
     if(r2 == 0) {
         char tmp[PATH_MAX];
-        strcpy(tmp, editing_dir);
-        strcat(tmp, file);
+        strcpy(tmp, file);
         string_obj* tmp2 = STRING_NEW("");
         saphire_get_quoted_fname_readline_ver(tmp, tmp2);
         char* ret = strdup(string_c_str(tmp2));
@@ -292,12 +290,122 @@ char* readline_file_completion_migemo_match(char* editing_dir, char* text, char*
     return NULL;
 }
 
+static BOOL read_one_argument(char* p, string_obj* buf)
+{
+    /// 単語の区切り記号かどうかのテーブル 127以下なら
+    static unsigned char table[] = { 
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
+        1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 
+        1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
+        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0 
+    };
 
+    BOOL squote = FALSE;
+    BOOL dquote = FALSE;
+    while(*p && (p - rl_line_buffer < rl_point)) {
+        /// 空文字
+        if(!squote && !dquote && *p == '"' && *(p+1) == '"') {
+            p+=2;
+            string_put(buf, "");
+        }
+        else if(!dquote && !squote && *p == '\'' && *(p+1) == '\'') {
+            p+=2;
+            string_put(buf, "");
+        }
+        else if(*p == '\\' && *(p+1) == 't') {
+            string_push_back2(buf, '\t');
+
+            p+=2;
+        }
+        else if(*p == '\\' && *(p+1) == 'n') {
+            string_push_back2(buf, '\n');
+
+            p+=2;
+        }
+        else if(*p == '\\' && *(p+1) == 'r') {
+            string_push_back2(buf, '\r');
+
+            p+=2;
+        }
+        else if(!squote && !dquote && *p == '\\') {
+            p++;
+
+            if(*p != 0) {
+                string_push_back2(buf, *p);
+                p++;
+            }
+            else {
+                break;
+            }
+        }
+        else if(!dquote && *p == '\'') {
+            p++;
+            squote = !squote;
+        }
+        else if(!squote && *p == '"') {
+            dquote = !dquote;
+            p++;
+        }
+        else if(squote || dquote) {
+            if(*p == 0) {
+                break;
+            }
+            else {
+                string_push_back2(buf, *p);
+
+                p++;
+            }
+        }
+        else if(((unsigned char)*p) > 127) {
+            string_push_back2(buf, *p);
+            p++;
+        }
+        else if(table[*p]) {
+            string_put(buf, "");
+            p++;
+        }
+        else {
+            string_push_back2(buf, *p);
+            p++;
+        }
+    }
+}
+
+void parentname2(char* result, char* path)
+{
+    char* p;
+
+    if(*path == 0) {
+        strcpy(result, "");
+        return;
+    }
+    
+    if(strcmp(path, "/") == 0) {
+        strcpy(result, "/");
+        return;
+    }
+
+    for(p=path + strlen(path)-1; p != path-1; p--) {
+        if(*p == '/' ) {
+            memcpy(result, path, p-path);
+            result[p-path] = '/';
+            result[p-path+1] = 0;
+            
+            return;
+        }
+    }
+
+    strcpy(result, "");
+}
+
+static string_obj* gEditingDir;
 
 char* readline_file_completion(const char* text, int rstat)
 {
-//printf("text (%s)\n", text);
-
     /// クォートを除外 ///
     char* text2 = MALLOC(strlen(text) +1);
     char* _p = (char*)text;
@@ -313,49 +421,66 @@ char* readline_file_completion(const char* text, int rstat)
     }
     *_p2 = 0;
 
-    static int index, wordlen, wordlen_editting_file;
-    static char editing_dir[PATH_MAX];
-    static char editing_dir2[PATH_MAX];
-    static char editing_file[PATH_MAX];
-    static int wordlen_editing_file;
+    static int index, wordlen;
 
     /// 初期化 ///
     if(rstat == 0) {
         wordlen = strlen(text2);
         index = 0;
 
-        /// ディレクトリのパートとファイルのパートに分ける ///
-        if(strcmp(text2, "") == 0) {
-            strcpy(editing_dir, "");
-            strcpy(editing_file, "");
-        }
-        else if(strstr(text2, "/") == NULL) {
-            strcpy(editing_dir, "");
-            strcpy(editing_file, text2);
-        }
-        else if(text2[strlen(text2) -1] == '/') {
-            strcpy(editing_dir, (char*)text2);
-            strcpy(editing_file, "");
-        }
-        else {
-            parentname(editing_dir, (char*)text2);
-            char tmp[PATH_MAX];   // basenameの引数は直接渡すとまずい
-            strcpy(tmp, (char*)text2);
-            strcpy(editing_file, basename(tmp));
-        }
-
-        wordlen_editing_file = strlen(editing_file);
+        // ディレクトリのパートを得る
+        string_put(gEditingDir, "");
+        read_one_argument(rl_line_buffer, gEditingDir);
+        char* tmp = MALLOC(string_length(gEditingDir) + 1);
+        parentname2(tmp, string_c_str(gEditingDir));
+        string_put(gEditingDir, tmp);
+        FREE(tmp);
 
         /// チルダから始まっている場合 ///
-        if(editing_dir[0] == '~') {
-            char* p = editing_dir;
+        if(string_c_str(gEditingDir)[0] == '~') {
+            char* p = string_c_str(gEditingDir);
             p++;
+            string_obj* user = STRING_NEW("");
+            while(*p && *p != '/') {
+                string_push_back2(user, *p);
+                p++;
+            }
 
-            strcpy(editing_dir2, getenv("HOME"));
-            strcat(editing_dir2, p);
-        }
-        else {
-            strcpy(editing_dir2, editing_dir);
+            /// ユーザー名からホームディレクトリに展開
+            string_obj* gEditingDir2 = STRING_NEW("");
+
+            if(string_length(user) == 0) {
+                char* home = getenv("HOME");
+                if(home) {
+                    string_put(gEditingDir2, home);
+                }
+                else {
+                    struct passwd* pw = getpwuid(getuid());
+                    if(pw) {
+                        string_put(gEditingDir2, pw->pw_dir);
+                    }
+                    else {
+                        string_put(gEditingDir2, "~");
+                    }
+                }
+            }
+            else {
+                struct passwd* pw = getpwnam(string_c_str(user));
+                if(pw) {
+                    string_put(gEditingDir2, pw->pw_dir);
+                }
+                else {
+                    string_put(gEditingDir2, "~");
+                    string_push_back(gEditingDir2, string_c_str(user));
+                }
+            }
+
+            string_push_back(gEditingDir2, p);
+
+            string_delete(user);
+
+            string_put(gEditingDir, string_c_str(gEditingDir2));
+            string_delete(gEditingDir2);
         }
 
         /// 補完候補を初期化 ///
@@ -371,8 +496,8 @@ char* readline_file_completion(const char* text, int rstat)
         }
 
         /// 補完候補をディレクトリから読み込み ///
-        if(strcmp(editing_dir2, "") != 0) {
-            DIR* dir = opendir(editing_dir2);
+        if(strcmp(string_c_str(gEditingDir), "") != 0) {
+            DIR* dir = opendir(string_c_str(gEditingDir));
 
             if(dir) {
                 struct dirent* entry;
@@ -383,7 +508,7 @@ char* readline_file_completion(const char* text, int rstat)
                         char buf[PATH_MAX];
 
                         struct stat _stat;
-                        strcpy(buf, editing_dir2);
+                        strcpy(buf, string_c_str(gEditingDir));
                         strcat(buf, entry->d_name);
                         stat(buf, &_stat);
 
@@ -437,11 +562,9 @@ char* readline_file_completion(const char* text, int rstat)
         index++;
 
         /// ファイル名の入力がない補完は判定しなくても全てマッチ ///
-        if(strcmp(editing_file, "") == 0)
-        {
+        if(strcmp(text2, "") == 0) {
             char path[PATH_MAX];
-            strcpy(path, editing_dir2);
-            strcat(path, file);
+            strcpy(path, file);
 
             string_obj* tmp2 = STRING_NEW("");
             saphire_get_quoted_fname_readline_ver(path, tmp2);
@@ -449,6 +572,8 @@ char* readline_file_completion(const char* text, int rstat)
             string_delete(tmp2);
 
             struct stat stat_;
+            strcpy(path, string_c_str(gEditingDir));
+            strcat(path, file);
             if(stat(path, &stat_) == 0) {
                 if(S_ISDIR(stat_.st_mode)) {
                     rl_completion_append_character= 0;
@@ -461,22 +586,25 @@ char* readline_file_completion(const char* text, int rstat)
                 rl_completion_append_character= ' ';
             }
 
-            FREE(text2);
-            if(ret) return ret;
+            if(ret) {
+                FREE(text2);
+                return ret;
+            }
         }
 
         /// アスキー文字同士の補完 ///
-        else if(is_all_ascii((char*)editing_file) && is_all_ascii(file)) {
-            if(!strncmp(editing_file, file, wordlen_editing_file)) {
+        else if(is_all_ascii((char*)text2) && is_all_ascii(file)) {
+            if(!strncmp(text2, file, wordlen)) {
                 char path[PATH_MAX];
-                strcpy(path, editing_dir2);
-                strcat(path, file);
+                strcpy(path, file);
 
                 string_obj* tmp2 = STRING_NEW("");
                 saphire_get_quoted_fname(path, tmp2);
                 char* ret = strdup(string_c_str(tmp2));
                 string_delete(tmp2);
 
+                strcpy(path, string_c_str(gEditingDir));
+                strcat(path, file);
                 struct stat stat_;
                 if(stat(path, &stat_) == 0) {
                     if(S_ISDIR(stat_.st_mode)) {
@@ -490,19 +618,19 @@ char* readline_file_completion(const char* text, int rstat)
                     rl_completion_append_character= ' ';
                 }
 
-                FREE(text2);
-
-                if(ret) return ret;
+                if(ret) { 
+                    FREE(text2);
+                    return ret; 
+                }
             }
         }
-
         /// migemo 補完 ///
         else {
-            char* ret = readline_file_completion_migemo_match(editing_dir, (char*)text2 , file);
+            char* ret = readline_file_completion_migemo_match((char*)text2 , file);
             if(ret) {
                 struct stat stat_;
                 char path[PATH_MAX];
-                strcpy(path, editing_dir2);
+                strcpy(path, string_c_str(gEditingDir));
                 strcat(path, file);
                 if(stat(path, &stat_) == 0) {
                     if(S_ISDIR(stat_.st_mode)) {
@@ -522,6 +650,14 @@ char* readline_file_completion(const char* text, int rstat)
         }
     }
 
+    if(index == 0 && vector_size(files) == 0) {
+        index ++;
+        FREE(text2);
+        rl_completion_append_character= 0;
+        return strdup("");
+    }
+
+
     FREE(text2);
     return NULL;
 }
@@ -715,6 +851,7 @@ char** readline_on_complete(const char* text, int start, int end)
     
     BOOL program = TRUE;
     BOOL env = FALSE;
+    BOOL user = FALSE;
 
     char* p = rl_line_buffer;
 
@@ -729,18 +866,22 @@ char** readline_on_complete(const char* text, int start, int end)
         if(*p == '\\' && *(p+1) == 't') {
             p+=2;
             env = FALSE;
+            user = FALSE;
         }
         else if(*p == '\\' && *(p+1) == 'n') {
             p+=2;
             env = FALSE;
+            user = FALSE;
         }
         else if(*p == '\\' && *(p+1) == 'r') {
             p+=2;
             env = FALSE;
+            user = FALSE;
         }
         else if(*p == '\\') {
             p+=2;
             env = FALSE;
+            user = FALSE;
         }
 
         // シングルクォート
@@ -748,12 +889,14 @@ char** readline_on_complete(const char* text, int start, int end)
             p++;
             squote = !squote;
             env = FALSE;
+            user = FALSE;
         }
         // ダブルクォート
         else if(!squote && *p == '"') {
             p++;
             dquote = !dquote;
             env = FALSE;
+            user = FALSE;
         }
         // 環境変数 $$()
         else if(!squote && *p == '$' && *(p+1) == '$' && *(p+2) == '(') {
@@ -765,6 +908,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         // 環境変数 $()
         else if(!squote && *p == '$' && *(p+1) == '(') {
@@ -775,6 +919,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         else if(!dquote && !squote && *p == ')') {
             p++;
@@ -783,33 +928,43 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
 
         // 環境変数 $$
         else if(!squote && *p == '$' && *(p+1) == '$') {
             p+=2;
             env = TRUE;
+            user = FALSE;
         }
         // 環境変数 $
         else if(!squote && *p == '$') {
             p+=2;
             env = TRUE;
+            user = FALSE;
         }
         /// シングルクォート、ダブルクォート中 ///
         else if(squote || dquote) {
             p++;
         }
+        else if(*p == '~') {
+            p++;
+            env = FALSE;
+            user = TRUE;
+        }
         /// グロブ ///
         else if(*p == '*' || *p == '?' || *p =='[' && (*(p+1) != ' ' && *(p+1) != '\t' && *(p+1) == '\n')) 
         {
             p++;
             env = FALSE;
+            user = FALSE;
         }
         else if(*p == '<' && program) {
             p++;
 
             program = FALSE;
             env = FALSE;
+            user = FALSE;
         }
         /// 単語の区切り /////////////////////////////////////////////////
     
@@ -821,6 +976,7 @@ char** readline_on_complete(const char* text, int start, int end)
 
             if(program) program = FALSE;
             env = FALSE;
+            user = FALSE;
         }
         /// ブロック
         else if(*p == ':' || *p == '{') {
@@ -830,6 +986,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         
         /// コマンドの区切り ///////////////////////////////////////////
@@ -843,6 +1000,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         
         /// パイプ ///
@@ -854,6 +1012,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         
         /// 文(statment)の区切り系 ///////////////////////////////////
@@ -867,6 +1026,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         /// オアオア ///
         else if(*p == '|' && *(p+1) == '|') {
@@ -877,6 +1037,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         /// バックグラウンド ///
         else if(*p == '&' && *(p+1) != '&') {
@@ -887,6 +1048,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         /// セミコロン ///
         else if(*p == ';') {
@@ -897,6 +1059,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         /// 改行 ///
         else if(*p == '\n') {
@@ -907,6 +1070,7 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
         }
         /// コメント ///
         else if(*p == '#') {
@@ -926,6 +1090,12 @@ char** readline_on_complete(const char* text, int start, int end)
                 p++;
             }
             env = FALSE;
+            user = FALSE;
+        }
+        else if(*p == '/') {
+            env = FALSE;
+            user = FALSE;
+            p++;
         }
         /// 終了 ///
         else if(*p == 0) {
@@ -939,6 +1109,7 @@ char** readline_on_complete(const char* text, int start, int end)
             p++;
             program = TRUE;
             env = FALSE;
+            user = FALSE;
             while(*p == ' ' || *p == '\t' || *p == '\n') {
                 p++;
             }
@@ -948,7 +1119,11 @@ char** readline_on_complete(const char* text, int start, int end)
         }
     }
 
-    if(env) {
+    if(user) {
+        rl_completion_append_character= 0;
+        return rl_completion_matches(text, rl_username_completion_function);
+    }
+    else if(env) {
         rl_completion_append_character= ' ';
         return rl_completion_matches(text, readline_env_completion);
     }
@@ -1591,8 +1766,8 @@ void readline_init()
 {
     rl_attempted_completion_function = readline_on_complete;
     rl_completer_quote_characters = "\"'";
-    rl_completer_word_break_characters = " \t\n\"'|{}!&;()<>=";
-    rl_filename_quote_characters = " \\\t\n\"|$*?[]!<>{}&;#()'";
+    rl_completer_word_break_characters = " \t\n\"'|{}!&;()<>=%/";
+    rl_filename_quote_characters = " \\\t\n\"|$*?[]!<>{}&;#()'%";
     rl_completion_append_character= ' ';
     rl_char_is_quoted_p = (rl_linebuf_func_t*)lftp_char_is_quoted;
     rl_filename_quoting_function = bash_quote_filename;
@@ -1603,10 +1778,12 @@ void readline_init()
 #endif
 
     gCompList = VECTOR_NEW(100);
+    gEditingDir = STRING_NEW("");
 }
 
 void readline_final()
 {
+    string_delete(gEditingDir);
     int i;
     for(i=0; i<vector_size(gCompList); i++) {
         string_delete(vector_item(gCompList, i));
