@@ -1961,12 +1961,24 @@ int mis_raw_mode()
 
 void msave_screen()
 {
+#if defined(__CYGWIN__)
+    char buf[128];
+    snprintf(buf, 128, "\033%d\033[?47h", 7);
+    ttywrite(buf);
+#else
     ttywrite(tigetstr("smcup"));
+#endif
 }
 
 void mrestore_screen()
 {
+#if defined(__CYGWIN__)
+    char buf[128];
+    snprintf(buf, 128, "\033[2J\033[?47l\033%d", 8);
+    ttywrite(buf);
+#else
     ttywrite(tigetstr("rmcup"));
+#endif
 }
 
 
