@@ -52,6 +52,8 @@ enum eCommandKind {
     kSweep,
     kShow,
     kYeild,
+    kEach,
+    kForeach,
 //    kParse,
     kCommand 
 };
