@@ -87,6 +87,8 @@ char* gStatmentKindStrs[kCommand+1] = {
     "sweep",
     "show",
     "yeild",
+    "each",
+    "foreach",
 //    "parse",
     "command"
 };
@@ -2541,7 +2543,7 @@ void saphire_init(enum eAppType app_type, BOOL job_control, enum eRuntimeScript
     gFileHandles = HASH_NEW(100);
     gReadBuffers = HASH_NEW(100);
 
-    setenv("SAPHIRE_VERSION", "1.0.5", 1);
+    setenv("SAPHIRE_VERSION", "1.0.6", 1);
     setenv("SAPHIRE_PROMPT", "> ", 1);
     setenv("SAPHIRE_DOCDIR", DOCDIR, 1);
     char* home = getenv("HOME");
